# Generated by Django 4.1.7 on 2024-02-29 11:34
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("datasets", "0017_dataset_enable_export"),
    ]

    operations = [
        migrations.AddField(
            model_name="datasettable",
            name="id_field",
            field=models.CharField(default="id", max_length=50),
        ),
    ]
