# Generated by Django 4.2.17 on 2025-01-29 10:47
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0018_datasettable_id_field"),
    ]

    operations = [
        migrations.AddField(
            model_name="datasettable",
            name="enable_export",
            field=models.BooleanField(default=False),
        ),
    ]
